@@ -1,4 +1,5 @@
-/* vim: set sw=4 sts=4: -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 ts=4 expandtab: */
 /*
    rsvg.c: SAX-based renderer for SVG files into a GdkPixbuf.
 
@@ -25,31 +26,38 @@
 
 #include "config.h"
 
-#ifdef HAVE_SVGZ
+#include "rsvg.h"
+#include "rsvg-private.h"
+#include "rsvg-css.h"
+#include "rsvg-styles.h"
+#include "rsvg-shapes.h"
+#include "rsvg-structure.h"
+#include "rsvg-image.h"
+#include "rsvg-text.h"
+#include "rsvg-filter.h"
+#include "rsvg-mask.h"
+#include "rsvg-marker.h"
+#include "rsvg-cairo-render.h"
+
+#include <libxml/uri.h>
+#include <libxml/parser.h>
+#include <libxml/parserInternals.h>
+
+#include <math.h>
+#include <string.h>
+#include <stdarg.h>
+
+#ifdef HAVE_GSF
 #include <gsf/gsf-input-gzip.h>
 #include <gsf/gsf-input-memory.h>
 #include <gsf/gsf-output-memory.h>
 #include <gsf/gsf-utils.h>
 #endif
 
-#include "rsvg.h"
-#include "rsvg-private.h"
-#include "rsvg-css.h"
-#include "rsvg-styles.h"
-#include "rsvg-shapes.h"
-#include "rsvg-image.h"
-#include "rsvg-text.h"
-#include "rsvg-filter.h"
-#include "rsvg-mask.h"
-#include "rsvg-marker.h"
-
-#include <math.h>
-#include <string.h>
-#include <stdarg.h>
-
 #include "rsvg-bpath-util.h"
 #include "rsvg-path.h"
 #include "rsvg-paint-server.h"
+#include "rsvg-xml.h"
 
 /*
  * This is configurable at runtime
@@ -259,7 +267,7 @@ rsvg_standard_element_start (RsvgHandle * ctx, const char *name, RsvgPropertyBag
 
     if (newnode) {
         newnode->type = g_string_new (name);
-	newnode->parent = ctx->priv->currentnode;
+        newnode->parent = ctx->priv->currentnode;
         rsvg_node_set_atts (newnode, ctx, atts);
         rsvg_defs_register_memory (ctx->priv->defs, newnode);
         if (ctx->priv->currentnode) {
@@ -685,6 +693,8 @@ rsvg_end_element (void *data, const xmlChar * name)
             && !strcmp ((const char *) name, ctx->priv->currentnode->type->str))
             rsvg_pop_def_group (ctx);
 
+        if (ctx->priv->treebase && !strcmp ((const char *)name, "svg"))
+            _rsvg_node_svg_apply_atts ((RsvgNodeSvg *)ctx->priv->treebase, ctx);
     }
 }
 
@@ -704,37 +714,35 @@ rsvg_characters_impl (RsvgHandle * ctx, const xmlChar * ch, int len)
     if (!ch || !len)
         return;
 
-	if (ctx->priv->currentnode)
-		{
-			if (!strcmp ("tspan", ctx->priv->currentnode->type->str) ||
-				!strcmp ("text", ctx->priv->currentnode->type->str))
-				{
-					guint i;
+    if (ctx->priv->currentnode) {
+        if (!strcmp ("tspan", ctx->priv->currentnode->type->str) ||
+            !strcmp ("text", ctx->priv->currentnode->type->str)) {
+            guint i;
 
-					/* find the last CHARS node in the text or tspan node, so that we
-					   can coalesce the text, and thus avoid screwing up the Pango layouts */
-					self = NULL;
-					for (i = 0; i < ctx->priv->currentnode->children->len; i++) {
-						RsvgNode *node = g_ptr_array_index (ctx->priv->currentnode->children, i);
-						if (!strcmp (node->type->str, "RSVG_NODE_CHARS")) {
-							self = (RsvgNodeChars*)node;
-						}
-					}
+            /* find the last CHARS node in the text or tspan node, so that we
+               can coalesce the text, and thus avoid screwing up the Pango layouts */
+            self = NULL;
+            for (i = 0; i < ctx->priv->currentnode->children->len; i++) {
+                RsvgNode *node = g_ptr_array_index (ctx->priv->currentnode->children, i);
+                if (!strcmp (node->type->str, "RSVG_NODE_CHARS")) {
+                    self = (RsvgNodeChars*)node;
+                }
+            }
 
-					if (self != NULL) {
-						if (!g_utf8_validate ((char *) ch, len, NULL)) {
-							char *utf8;
-							utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
-							g_string_append (self->contents, utf8);
-							g_free (utf8);
-						} else {
-							g_string_append_len (self->contents, (char *)ch, len);
-						}
+            if (self != NULL) {
+                if (!g_utf8_validate ((char *) ch, len, NULL)) {
+                    char *utf8;
+                    utf8 = rsvg_make_valid_utf8 ((char *) ch, len);
+                    g_string_append (self->contents, utf8);
+                    g_free (utf8);
+                } else {
+                    g_string_append_len (self->contents, (char *)ch, len);
+                }
 
-						return;
-					}
-				}
-		}
+                return;
+            }
+        }
+    }
 
     self = g_new (RsvgNodeChars, 1);
     _rsvg_node_init (&self->super);
@@ -796,6 +804,36 @@ rsvg_entity_decl (void *data, const xmlChar * name, int type,
     RsvgHandle *ctx = (RsvgHandle *) data;
     GHashTable *entities = ctx->priv->entities;
     xmlEntityPtr entity;
+#if LIBXML_VERSION >= 20700
+    xmlChar *resolvedSystemId = NULL, *resolvedPublicId = NULL;
+
+    if (systemId)
+        resolvedSystemId = xmlBuildRelativeURI (systemId, (xmlChar*) rsvg_handle_get_base_uri (ctx));
+    else if (publicId)
+        resolvedPublicId = xmlBuildRelativeURI (publicId, (xmlChar*) rsvg_handle_get_base_uri (ctx));
+
+    if (type == XML_EXTERNAL_PARAMETER_ENTITY && !content) {
+      GByteArray *arr = NULL;
+
+      if (systemId)
+        arr = _rsvg_acquire_xlink_href_resource ((const char *) systemId,
+                                                 rsvg_handle_get_base_uri (ctx), NULL);
+      else if (publicId)
+        arr = _rsvg_acquire_xlink_href_resource ((const char *) publicId,
+                                                 rsvg_handle_get_base_uri (ctx), NULL);
+      if (arr) {
+        content = xmlCharStrndup ((const char*)arr->data, arr->len);
+        g_byte_array_free(arr, TRUE);
+      }
+    }
+      
+    entity = xmlNewEntity(NULL, name, type, resolvedPublicId, resolvedSystemId, content);
+
+    xmlFree(resolvedPublicId);
+    xmlFree(resolvedSystemId);
+
+    g_hash_table_insert (entities, g_strdup ((const char*) name), entity);
+#else
     xmlChar *dupname;
 
     entity = xmlMalloc (sizeof (xmlEntity));
@@ -834,6 +872,7 @@ rsvg_entity_decl (void *data, const xmlChar * name, int type,
     }
 
     g_hash_table_insert (entities, dupname, entity);
+#endif
 }
 
 static void
@@ -845,6 +884,17 @@ rsvg_unparsed_entity_decl (void *ctx,
     rsvg_entity_decl (ctx, name, XML_INTERNAL_GENERAL_ENTITY, publicId, systemId, NULL);
 }
 
+static xmlEntityPtr
+rsvg_get_parameter_entity (void *data, const xmlChar * name)
+{
+    RsvgHandle *ctx = (RsvgHandle *) data;
+    xmlEntityPtr entity;
+
+    entity = g_hash_table_lookup (ctx->priv->entities, name);
+
+    return entity;
+}
+
 #endif
 
 static void
@@ -913,6 +963,7 @@ rsvg_SAX_handler_struct_init (void)
         rsvgSAXHandlerStruct.getEntity = rsvg_get_entity;
         rsvgSAXHandlerStruct.entityDecl = rsvg_entity_decl;
         rsvgSAXHandlerStruct.unparsedEntityDecl = rsvg_unparsed_entity_decl;
+        rsvgSAXHandlerStruct.getParameterEntity = rsvg_get_parameter_entity;
 #endif
         rsvgSAXHandlerStruct.characters = rsvg_characters;
         rsvgSAXHandlerStruct.error = rsvg_error_cb;
@@ -931,26 +982,27 @@ rsvg_path_is_uri (char const *path)
     char const *p;
 
     if (path == NULL)
-	return FALSE;
+        return FALSE;
 
     if (strlen (path) < 4)
-	return FALSE;
+        return FALSE;
 
-    if (   (path[0] < 'a' || path[0] > 'z')
-	&& (path[0] < 'A' || path[0] > 'Z'))
-	return FALSE;
+    if ((path[0] < 'a' || path[0] > 'z') &&
+        (path[0] < 'A' || path[0] > 'Z')) {
+        return FALSE;
+    }
 
     for (p = &path[1];
-	    (*p >= 'a' && *p <= 'z') 
-	 || (*p >= 'A' && *p <= 'Z') 
-	 || (*p >= '0' && *p <= '9') 
-	 || *p == '+' 
-	 || *p == '-' 
-	 || *p == '.';
-	 p++);
+	    (*p >= 'a' && *p <= 'z') ||
+        (*p >= 'A' && *p <= 'Z') ||
+        (*p >= '0' && *p <= '9') ||
+         *p == '+' ||
+         *p == '-' ||
+         *p == '.';
+        p++);
 
     if (strlen (p) < 3)
-	return FALSE;
+        return FALSE;
 
     return (p[0] == ':' && p[1] == '/' && p[2] == '/');
 }
@@ -996,9 +1048,9 @@ rsvg_handle_set_base_uri (RsvgHandle * handle, const char *base_uri)
 	return;
 
     if (rsvg_path_is_uri (base_uri)) 
-	uri = g_strdup (base_uri);
+        uri = g_strdup (base_uri);
     else
-	uri = rsvg_get_base_uri_from_filename (base_uri);
+        uri = rsvg_get_base_uri_from_filename (base_uri);
 
     if (uri) {
         if (handle->priv->base_uri)
@@ -1008,6 +1060,39 @@ rsvg_handle_set_base_uri (RsvgHandle * handle, const char *base_uri)
     }
 }
 
+/**
+ * rsvg_handle_set_base_gfile:
+ * @handle: a #RsvgHandle
+ * @file: a #GFile
+ *
+ * Set the base URI for @handle from @file.
+ * Note: This function may only be called before rsvg_handle_write()
+ * or rsvg_handle_read_stream_sync() has been called.
+ *
+ * Since: 2.32
+ */
+void
+rsvg_handle_set_base_gfile (RsvgHandle *handle,
+                            GFile      *base_file)
+{
+    RsvgHandlePrivate *priv;
+
+    g_return_if_fail (RSVG_IS_HANDLE (handle));
+    g_return_if_fail (G_IS_FILE (base_file));
+
+    priv = handle->priv;
+
+    g_object_ref (base_file);
+    if (priv->base_gfile)
+        g_object_unref (priv->base_gfile);
+    priv->base_gfile = base_file;
+    
+    g_free (priv->base_uri);
+    priv->base_uri = g_file_get_uri (base_file);
+
+    rsvg_defs_set_base_uri (priv->defs, priv->base_uri);
+}
+
 /**
  * rsvg_handle_get_base_uri:
  * @handle: A #RsvgHandle
@@ -1015,7 +1100,7 @@ rsvg_handle_set_base_uri (RsvgHandle * handle, const char *base_uri)
  * Gets the base uri for this #RsvgHandle.
  *
  * Returns: the base uri, possibly null
- * Since: 2.9 (really present in 2.8 as well)
+ * Since: 2.8
  */
 G_CONST_RETURN char *
 rsvg_handle_get_base_uri (RsvgHandle * handle)
@@ -1039,6 +1124,24 @@ rsvg_error_quark (void)
     return g_quark_from_string ("rsvg-error-quark");
 }
 
+static void
+rsvg_set_error (GError **error, xmlParserCtxtPtr ctxt)
+{
+    xmlErrorPtr xerr;
+
+    xerr = xmlCtxtGetLastError (ctxt);
+    if (xerr) {
+        g_set_error (error, rsvg_error_quark (), 0,
+                     _("Error domain %d code %d on line %d column %d of %s: %s"),
+                     xerr->domain, xerr->code,
+                     xerr->line, xerr->int2,
+                     xerr->file ? xerr->file : "data",
+                     xerr->message ? xerr->message: "-");
+    } else {
+        g_set_error (error, rsvg_error_quark (), 0, _("Error parsing XML data"));
+    }
+}
+
 static gboolean
 rsvg_handle_write_impl (RsvgHandle * handle, const guchar * buf, gsize count, GError ** error)
 {
@@ -1060,7 +1163,7 @@ rsvg_handle_write_impl (RsvgHandle * handle, const guchar * buf, gsize count, GE
 
     result = xmlParseChunk (handle->priv->ctxt, (char *) buf, count, 0);
     if (result != 0) {
-        g_set_error (error, rsvg_error_quark (), 0, _("Error parsing XML data"));
+        rsvg_set_error (error, handle->priv->ctxt);
         return FALSE;
     }
 
@@ -1094,7 +1197,7 @@ rsvg_handle_close_impl (RsvgHandle * handle, GError ** error)
         xmlFreeDoc (xmlDoc);
 
         if (result != 0) {
-            g_set_error (error, rsvg_error_quark (), 0, _("Error parsing XML data"));
+            rsvg_set_error (error, handle->priv->ctxt);
             return FALSE;
         }
     }
@@ -1111,24 +1214,18 @@ rsvg_handle_close_impl (RsvgHandle * handle, GError ** error)
     return TRUE;
 }
 
-static void
-rsvg_state_free_func (gpointer data, gpointer user_data)
-{
-    rsvg_state_finalize ((RsvgState *) data);
-    g_slice_free (RsvgState, data);
-}
-
 void
 rsvg_drawing_ctx_free (RsvgDrawingCtx * handle)
 {
     rsvg_render_free (handle->render);
 
-    g_slist_foreach (handle->state, rsvg_state_free_func, (gpointer) handle);
-    g_slist_free (handle->state);
+    rsvg_state_free_all (handle->state);
 
 	/* the drawsub stack's nodes are owned by the ->defs */
 	g_slist_free (handle->drawsub_stack);
 
+    g_slist_free (handle->ptrs);
+	
     if (handle->base_uri)
         g_free (handle->base_uri);
 
@@ -1204,131 +1301,6 @@ rsvg_handle_get_desc (RsvgHandle * handle)
         return NULL;
 }
 
-typedef struct {
-    RsvgRender super;
-    RsvgBbox bbox;
-} RsvgBboxRender;
-
-static void
-rsvg_bbox_render_path (RsvgDrawingCtx * ctx, const RsvgBpathDef * bpath_def)
-{
-    RsvgState *state = rsvg_state_current (ctx);
-    RsvgBpath *bpath;
-    RsvgBboxRender *render = (RsvgBboxRender *) ctx->render;
-    RsvgBbox bbox;
-    int i;
-
-    rsvg_bbox_init (&bbox, state->affine);
-    bbox.w = bbox.h = bbox.virgin = 0;
-
-    for (i = 0; i < bpath_def->n_bpath; i++) {
-        bpath = &bpath_def->bpath[i];
-
-        switch (bpath->code) {
-        case RSVG_MOVETO:
-        case RSVG_MOVETO_OPEN:
-        case RSVG_CURVETO:
-        case RSVG_LINETO:
-            bbox.x = bpath->x3;
-            bbox.y = bpath->y3;
-            rsvg_bbox_insert (&render->bbox, &bbox);
-            break;
-        default:
-            break;
-        }
-    }
-}
-
-static void
-rsvg_bbox_render_image (RsvgDrawingCtx * ctx,
-                        const GdkPixbuf * pixbuf,
-                        double pixbuf_x, double pixbuf_y, double w, double h)
-{
-    RsvgState *state = rsvg_state_current (ctx);
-    RsvgBboxRender *render = (RsvgBboxRender *) ctx->render;
-    RsvgBbox bbox;
-
-    rsvg_bbox_init (&bbox, state->affine);
-    bbox.x = pixbuf_x;
-    bbox.y = pixbuf_y;
-    bbox.w = w;
-    bbox.h = h;
-    bbox.virgin = 0;
-
-    rsvg_bbox_insert (&render->bbox, &bbox);
-}
-
-
-static void
-rsvg_bbox_render_free (RsvgRender * self)
-{
-    g_free (self);
-}
-
-static void
-rsvg_bbox_push_discrete_layer (RsvgDrawingCtx * ctx)
-{
-}
-
-static void
-rsvg_bbox_pop_discrete_layer (RsvgDrawingCtx * ctx)
-{
-}
-
-static void
-rsvg_bbox_add_clipping_rect (RsvgDrawingCtx * ctx, double x, double y, double w, double h)
-{
-}
-
-static RsvgBboxRender *
-rsvg_bbox_render_new ()
-{
-    RsvgBboxRender *render = g_new0 (RsvgBboxRender, 1);
-    double affine[6];
-
-    render->super.free = rsvg_bbox_render_free;
-    render->super.render_image = rsvg_bbox_render_image;
-    render->super.render_path = rsvg_bbox_render_path;
-    render->super.pop_discrete_layer = rsvg_bbox_pop_discrete_layer;
-    render->super.push_discrete_layer = rsvg_bbox_push_discrete_layer;
-    render->super.add_clipping_rect = rsvg_bbox_add_clipping_rect;
-    render->super.get_image_of_node = NULL;
-    _rsvg_affine_identity (affine);
-    rsvg_bbox_init (&render->bbox, affine);
-
-    return render;
-}
-
-static RsvgBbox
-_rsvg_find_bbox (RsvgHandle * handle)
-{
-    RsvgDrawingCtx *ctx = g_new (RsvgDrawingCtx, 1);
-    RsvgBbox output;
-    RsvgBboxRender *render = rsvg_bbox_render_new ();
-    ctx->drawsub_stack = NULL;
-    ctx->render = (RsvgRender *) render;
-
-    ctx->state = NULL;
-
-    ctx->defs = handle->priv->defs;
-    ctx->base_uri = g_strdup (handle->priv->base_uri);
-    ctx->dpi_x = handle->priv->dpi_x;
-    ctx->dpi_y = handle->priv->dpi_y;
-    ctx->vb.w = 512;
-    ctx->vb.h = 512;
-    ctx->pango_context = NULL;
-
-    rsvg_state_push (ctx);
-    _rsvg_affine_identity (rsvg_state_current (ctx)->affine);
-    _rsvg_node_draw_children ((RsvgNode *) handle->priv->treebase, ctx, 0);
-    rsvg_state_pop (ctx);
-
-    output = render->bbox;
-    rsvg_render_free (ctx->render);
-    g_free (ctx);
-    return output;
-}
-
 /**
  * rsvg_handle_get_dimensions
  * @handle: A #RsvgHandle
@@ -1341,41 +1313,255 @@ _rsvg_find_bbox (RsvgHandle * handle)
 void
 rsvg_handle_get_dimensions (RsvgHandle * handle, RsvgDimensionData * dimension_data)
 {
-    RsvgNodeSvg *sself;
+    /* This function is probably called from the cairo_render functions.
+     * To prevent an infinite loop we are saving the state.
+     */
+    if (!handle->priv->in_loop) {
+        handle->priv->in_loop = TRUE;
+        rsvg_handle_get_dimensions_sub (handle, dimension_data, NULL);
+        handle->priv->in_loop = FALSE;
+    } else {
+        /* Called within the size function, so return a standard size */
+        dimension_data->em = dimension_data->width = 1;
+        dimension_data->ex = dimension_data->height = 1;
+    }
+}
+
+/**
+ * rsvg_handle_get_dimensions_sub
+ * @handle: A #RsvgHandle
+ * @dimension_data: A place to store the SVG's size
+ * @id: An element's id within the SVG, or NULL to get the dimension of the whole SVG. 
+ * For example, if you have a layer called "layer1" for that you want to get the dimension, 
+ * pass "#layer1" as the id.
+ *
+ * Get the size of a subelement of the SVG file. Do not call from within the size_func callback, because an infinite loop will occur.
+ *
+ * Since: 2.22
+ */
+gboolean
+rsvg_handle_get_dimensions_sub (RsvgHandle * handle, RsvgDimensionData * dimension_data, const char *id)
+{
+    cairo_t *cr;
+    cairo_surface_t *target;
+    RsvgDrawingCtx *draw;
+    RsvgCairoRender *render;
+    RsvgNodeSvg *root = NULL;
+    RsvgNode *sself = NULL;
     RsvgBbox bbox;
 
-    g_return_if_fail (handle);
-    g_return_if_fail (dimension_data);
+    gboolean handle_subelement = TRUE;
+
+    g_return_val_if_fail (handle, FALSE);
+    g_return_val_if_fail (dimension_data, FALSE);
 
     memset (dimension_data, 0, sizeof (RsvgDimensionData));
 
-    sself = (RsvgNodeSvg *) handle->priv->treebase;
-    if (!sself)
-        return;
+    if (id && *id) {
+        sself = (RsvgNode *) rsvg_defs_lookup (handle->priv->defs, id);
+
+        if (sself == (RsvgNode *) handle->priv->treebase)
+            id = NULL;
+    }
+    else
+        sself = (RsvgNode *) handle->priv->treebase;
+
+    if (!sself && id)
+        return FALSE;
+
+    root = (RsvgNodeSvg *) handle->priv->treebase;
+
+    if (!root)
+        return FALSE;
 
     bbox.x = bbox.y = 0;
     bbox.w = bbox.h = 1;
 
-    if (sself->w.factor == 'p' || sself->h.factor == 'p') {
-        if (sself->vbox.active && sself->vbox.w > 0. && sself->vbox.h > 0.) {
-            bbox.w = sself->vbox.w;
-            bbox.h = sself->vbox.h;
-        } else
-            bbox = _rsvg_find_bbox (handle);
+    if (!id && (root->w.factor == 'p' || root->h.factor == 'p')
+            && !root->vbox.active)
+        handle_subelement = TRUE;
+    else if (!id && root->w.length != -1 && root->h.length != -1)
+        handle_subelement = FALSE;
+
+    if (handle_subelement == TRUE) {
+        target = cairo_image_surface_create (CAIRO_FORMAT_RGB24,
+                                             1, 1);
+        cr = cairo_create  (target);
+
+        draw = rsvg_cairo_new_drawing_ctx (cr, handle);
+
+        if (!draw) {
+            cairo_destroy (cr);
+            cairo_surface_destroy (target);
+
+            return FALSE;
+        }
+
+        while (sself != NULL) {
+            draw->drawsub_stack = g_slist_prepend (draw->drawsub_stack, sself);
+            sself = sself->parent;
+        }
+
+        rsvg_state_push (draw);
+        cairo_save (cr);
+
+        rsvg_node_draw ((RsvgNode *) handle->priv->treebase, draw, 0);
+        render = (RsvgCairoRender *) draw->render;
+
+        bbox.x = render->bbox.x;
+        bbox.y = render->bbox.y;
+        bbox.w = render->bbox.w;
+        bbox.h = render->bbox.h;
+
+        cairo_restore (cr);
+        rsvg_state_pop (draw);
+        rsvg_drawing_ctx_free (draw);
+        cairo_destroy (cr);
+        cairo_surface_destroy (target);
+
+        dimension_data->width = bbox.w;
+        dimension_data->height = bbox.h;
+    } else {
+        bbox.w = root->vbox.w;
+        bbox.h = root->vbox.h;
+
+        dimension_data->width = (int) (_rsvg_css_hand_normalize_length (&root->w, handle->priv->dpi_x,
+                                       bbox.w + bbox.x * 2, 12) + 0.5);
+        dimension_data->height = (int) (_rsvg_css_hand_normalize_length (&root->h, handle->priv->dpi_y,
+                                         bbox.h + bbox.y * 2,
+                                         12) + 0.5);
     }
-
-    dimension_data->width = (int) (_rsvg_css_hand_normalize_length (&sself->w, handle->priv->dpi_x,
-                                                                    bbox.w + bbox.x * 2, 12) + 0.5);
-    dimension_data->height = (int) (_rsvg_css_hand_normalize_length (&sself->h, handle->priv->dpi_y,
-                                                                     bbox.h + bbox.y * 2,
-                                                                     12) + 0.5);
-
+    
     dimension_data->em = dimension_data->width;
     dimension_data->ex = dimension_data->height;
 
     if (handle->priv->size_func)
         (*handle->priv->size_func) (&dimension_data->width, &dimension_data->height,
                                     handle->priv->user_data);
+
+    return TRUE;
+}
+
+/**
+ * rsvg_handle_get_position_sub
+ * @handle: A #RsvgHandle
+ * @position_data: A place to store the SVG fragment's position.
+ * @id: An element's id within the SVG.
+ * For example, if you have a layer called "layer1" for that you want to get
+ * the position, pass "#layer1" as the id.
+ *
+ * Get the position of a subelement of the SVG file. Do not call from within
+ * the size_func callback, because an infinite loop will occur.
+ *
+ * Since: 2.22
+ */
+gboolean
+rsvg_handle_get_position_sub (RsvgHandle * handle, RsvgPositionData * position_data, const char *id)
+{
+    RsvgDrawingCtx		*draw;
+    RsvgNodeSvg			*root;
+    RsvgNode			*node;
+    RsvgBbox			 bbox;
+    RsvgDimensionData    dimension_data;
+    cairo_surface_t		*target = NULL;
+    cairo_t				*cr = NULL;
+    gboolean			 ret = FALSE;
+
+    g_return_val_if_fail (handle, FALSE);
+    g_return_val_if_fail (position_data, FALSE);
+
+    /* Short-cut when no id is given. */
+    if (NULL == id || '\0' == *id) {
+        position_data->x = 0;
+        position_data->y = 0;
+        return TRUE;
+    }
+
+    memset (position_data, 0, sizeof (*position_data));
+    memset (&dimension_data, 0, sizeof (dimension_data));
+
+    node = (RsvgNode *) rsvg_defs_lookup (handle->priv->defs, id);
+    if (!node) {
+        return FALSE;
+    } else if (node == (RsvgNode *) handle->priv->treebase) {
+        /* Root node. */
+        position_data->x = 0;
+        position_data->y = 0;
+        return TRUE;
+    }
+
+    root = (RsvgNodeSvg *) handle->priv->treebase;
+    if (!root)
+        return FALSE;
+
+    target = cairo_image_surface_create (CAIRO_FORMAT_RGB24, 1, 1);
+    cr = cairo_create  (target);
+    draw = rsvg_cairo_new_drawing_ctx (cr, handle);
+    if (!draw)
+        goto bail;
+
+    while (node != NULL) {
+        draw->drawsub_stack = g_slist_prepend (draw->drawsub_stack, node);
+        node = node->parent;
+    }
+
+    rsvg_state_push (draw);
+    cairo_save (cr);
+
+    rsvg_node_draw ((RsvgNode *) handle->priv->treebase, draw, 0);
+    bbox.x = ((RsvgCairoRender *) draw->render)->bbox.x;
+    bbox.y = ((RsvgCairoRender *) draw->render)->bbox.y;
+    bbox.w = ((RsvgCairoRender *) draw->render)->bbox.w;
+    bbox.h = ((RsvgCairoRender *) draw->render)->bbox.h;
+
+    cairo_restore (cr);
+    rsvg_state_pop (draw);
+    rsvg_drawing_ctx_free (draw);
+
+    position_data->x = bbox.x;
+    position_data->y = bbox.y;
+    dimension_data.width = bbox.w;
+    dimension_data.height = bbox.h;
+
+    dimension_data.em = dimension_data.width;
+    dimension_data.ex = dimension_data.height;
+
+    if (handle->priv->size_func)
+        (*handle->priv->size_func) (&dimension_data.width, &dimension_data.height,
+                                    handle->priv->user_data);
+
+    ret = TRUE;
+
+bail:
+    if (cr)
+        cairo_destroy (cr);
+    if (target)
+        cairo_surface_destroy (target);
+
+    return ret;
+}
+
+/** 
+ * rsvg_handle_has_sub:
+ * @handle: a #RsvgHandle
+ * @id: an element's id within the SVG
+ *
+ * Checks whether the element @id exists in the SVG document.
+ *
+ * Returns: %TRUE if @id exists in the SVG document
+ *
+ * Since: 2.22
+ */
+gboolean
+rsvg_handle_has_sub (RsvgHandle * handle,
+                     const char *id)
+{
+    g_return_val_if_fail (handle, FALSE);
+
+    if (G_UNLIKELY (!id || !id[0]))
+      return FALSE;
+
+    return rsvg_defs_lookup (handle->priv->defs, id) != NULL;
 }
 
 /** 
@@ -1496,79 +1682,109 @@ rsvg_handle_set_size_callback (RsvgHandle * handle,
 
 /**
  * rsvg_handle_write:
- * @handle: An #RsvgHandle
- * @buf: Pointer to svg data
+ * @handle: an #RsvgHandle
+ * @buf: (array length=count) (element-type uint8): pointer to svg data
  * @count: length of the @buf buffer in bytes
- * @error: return location for errors
+ * @error: (allow-none): a location to store a #GError, or %NULL
  *
  * Loads the next @count bytes of the image.  This will return #TRUE if the data
  * was loaded successful, and #FALSE if an error occurred.  In the latter case,
  * the loader will be closed, and will not accept further writes. If FALSE is
- * returned, @error will be set to an error from the #RSVG_ERROR domain.
+ * returned, @error will be set to an error from the #RsvgError domain. Errors
+ * from #GIOErrorEnum are also possible.
  *
- * Returns: #TRUE if the write was successful, or #FALSE if there was an
- * error.
+ * Returns: %TRUE on success, or %FALSE on error
  **/
 gboolean
 rsvg_handle_write (RsvgHandle * handle, const guchar * buf, gsize count, GError ** error)
 {
-    rsvg_return_val_if_fail (handle, FALSE, error);
-    rsvg_return_val_if_fail (!handle->priv->is_closed, FALSE, error);
+    RsvgHandlePrivate *priv;
 
-    if (handle->priv->first_write) {
-        handle->priv->first_write = FALSE;
+    rsvg_return_val_if_fail (handle, FALSE, error);
+    priv = handle->priv;
+
+    rsvg_return_val_if_fail (!priv->is_closed, FALSE, error);
+
+    if (priv->first_write) {
+        priv->first_write = FALSE;
 
         /* test for GZ marker. todo: store the first 2 bytes in the odd circumstance that someone calls
          * write() in 1 byte increments */
         if ((count >= 2) && (buf[0] == (guchar) 0x1f) && (buf[1] == (guchar) 0x8b)) {
-            handle->priv->is_gzipped = TRUE;
-
-#ifdef HAVE_SVGZ
-            handle->priv->gzipped_data = GSF_OUTPUT (gsf_output_memory_new ());
+#if GLIB_CHECK_VERSION (2, 24, 0)
+            priv->data_input_stream = g_memory_input_stream_new ();
+#elif defined (HAVE_GSF)
+            priv->gzipped_data = GSF_OUTPUT (gsf_output_memory_new ());
+#else
+            g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+                                 "GZip compressed SVG not supported");
+            return FALSE;
 #endif
         }
     }
 
-    if (handle->priv->is_gzipped) {
-#ifdef HAVE_SVGZ
-        return gsf_output_write (handle->priv->gzipped_data, count, buf);
-#else
-        return FALSE;
-#endif
+#if GLIB_CHECK_VERSION (2, 24, 0)
+    if (priv->data_input_stream) {
+        g_memory_input_stream_add_data ((GMemoryInputStream *) priv->data_input_stream,
+                                        g_memdup (buf, count), count, (GDestroyNotify) g_free);
+        return TRUE;
     }
+#elif defined (HAVE_GSF)
+    if (priv->gzipped_data)
+        return gsf_output_write (handle->priv->gzipped_data, count, buf);
+#endif
 
     return rsvg_handle_write_impl (handle, buf, count, error);
 }
 
 /**
  * rsvg_handle_close:
- * @handle: A #RsvgHandle
- * @error: A #GError
+ * @handle: a #RsvgHandle
+ * @error: (allow-none): a location to store a #GError, or %NULL
  *
  * Closes @handle, to indicate that loading the image is complete.  This will
- * return #TRUE if the loader closed successfully.  Note that @handle isn't
+ * return %TRUE if the loader closed successfully.  Note that @handle isn't
  * freed until @g_object_unref is called.
  *
- * Returns: #TRUE if the loader closed successfully, or #FALSE if there was
- * an error.
+ * Returns: %TRUE on success, or %FALSE on error
  **/
 gboolean
 rsvg_handle_close (RsvgHandle * handle, GError ** error)
 {
+    RsvgHandlePrivate *priv;
+
     rsvg_return_val_if_fail (handle, FALSE, error);
+    priv = handle->priv;
 
-	if (handle->priv->is_closed)
-		return TRUE;
+    if (priv->is_closed)
+          return TRUE;
 
-#if HAVE_SVGZ
-    if (handle->priv->is_gzipped) {
+#if GLIB_CHECK_VERSION (2, 24, 0)
+    if (priv->data_input_stream) {
+        GConverter *converter;
+        GInputStream *stream;
+        gboolean ret;
+
+        converter = G_CONVERTER (g_zlib_decompressor_new (G_ZLIB_COMPRESSOR_FORMAT_GZIP));
+        stream = g_converter_input_stream_new (priv->data_input_stream, converter);
+        g_object_unref (converter);
+        g_object_unref (priv->data_input_stream);
+        priv->data_input_stream = NULL;
+
+        ret = rsvg_handle_read_stream_sync (handle, stream, NULL, error);
+        g_object_unref (stream);
+
+        return ret;
+    }
+#elif defined(HAVE_GSF)
+    if (priv->gzipped_data) {
         GsfInput *gzip;
         const guchar *bytes;
         gsize size;
         gsize remaining;
 
-        bytes = gsf_output_memory_get_bytes (GSF_OUTPUT_MEMORY (handle->priv->gzipped_data));
-        size = gsf_output_size (handle->priv->gzipped_data);
+        bytes = gsf_output_memory_get_bytes (GSF_OUTPUT_MEMORY (priv->gzipped_data));
+        size = gsf_output_size (priv->gzipped_data);
 
         gzip =
             GSF_INPUT (gsf_input_gzip_new
@@ -1595,20 +1811,198 @@ rsvg_handle_close (RsvgHandle * handle, GError ** error)
             }
             remaining = gsf_input_remaining (gzip);
         }
-        g_object_unref (G_OBJECT (gzip));
+        g_object_unref (gzip);
 
         /* close parent */
-        gsf_output_close (handle->priv->gzipped_data);
+        gsf_output_close (priv->gzipped_data);
+        g_object_unref (priv->gzipped_data);
+        priv->gzipped_data = NULL;
     }
-#endif
+#endif /* GIO >= 2.24.0 */
 
     return rsvg_handle_close_impl (handle, error);
 }
 
-#ifdef HAVE_GNOME_VFS
-#include <libgnomevfs/gnome-vfs.h>
+/**
+ * rsvg_handle_read_stream_sync:
+ * @handle: a #RsvgHandle
+ * @stream: a #GInputStream
+ * @cancellable: (allow-none): a #GCancellable, or %NULL
+ * @error: (allow-none): a location to store a #GError, or %NULL
+ *
+ * Reads @stream and writes the data from it to @handle.
+ *
+ * If @cancellable is not %NULL, then the operation can be cancelled by
+ * triggering the cancellable object from another thread. If the
+ * operation was cancelled, the error G_IO_ERROR_CANCELLED will be
+ * returned.
+ *
+ * Returns: %TRUE if reading @stream succeeded, or %FALSE otherwise
+ *   with @error filled in
+ *
+ * Since: 2.32
+ */
+gboolean
+rsvg_handle_read_stream_sync (RsvgHandle   *handle,
+                              GInputStream *stream,
+                              GCancellable *cancellable,
+                              GError      **error)
+{
+    RsvgHandlePrivate *priv;
+    xmlParserInputBufferPtr buffer;
+    xmlParserInputPtr input;
+    int result;
+    xmlDocPtr doc;
+    GError *err = NULL;
+
+    g_return_val_if_fail (RSVG_IS_HANDLE (handle), FALSE);
+    g_return_val_if_fail (G_IS_INPUT_STREAM (stream), FALSE);
+    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    priv = handle->priv;
+
+    priv->error = &err;
+    if (priv->ctxt == NULL) {
+        priv->ctxt = xmlCreatePushParserCtxt (&rsvgSAXHandlerStruct, handle, NULL, 0,
+                                              rsvg_handle_get_base_uri (handle));
+
+        /* if false, external entities work, but internal ones don't. if true, internal entities
+           work, but external ones don't. favor internal entities, in order to not cause a
+           regression */
+        /* FIXMEchpe: FIX THIS! */
+        priv->ctxt->replaceEntities = TRUE;
+    }
+
+    buffer = _rsvg_xml_input_buffer_new_from_stream (stream, cancellable, XML_CHAR_ENCODING_NONE, &err);
+    input = xmlNewIOInputStream (priv->ctxt, buffer, XML_CHAR_ENCODING_NONE);
+#if LIBXML_VERSION >= 20700
+    if (xmlPushInput (priv->ctxt, input) < 0) {
+        rsvg_set_error (error, priv->ctxt);
+        xmlFreeInputStream (input);
+        return FALSE;
+    }
+#else
+    xmlPushInput (priv->ctxt, input);
 #endif
 
+    result = xmlParseDocument (priv->ctxt);
+    if (result != 0) {
+        if (err)
+            g_propagate_error (error, err);
+        else
+            rsvg_set_error (error, handle->priv->ctxt);
+
+        return FALSE;
+    }
+
+    priv->error = NULL;
+
+    if (err != NULL) {
+        g_propagate_error (error, err);
+        return FALSE;
+    }
+
+    doc = priv->ctxt->myDoc;
+    xmlFreeParserCtxt (priv->ctxt);
+    priv->ctxt = NULL;
+
+    xmlFreeDoc (doc);
+
+    rsvg_defs_resolve_all (priv->defs);
+    priv->finished = TRUE;
+
+    return TRUE;
+}
+
+/**
+ * rsvg_handle_new_from_gfile_sync:
+ * @file: a #GFile
+ * @flags: flags from #RsvgHandleFlags
+ * @cancellable: (allow-none): a #GCancellable, or %NULL
+ * @error: (allow-none): a location to store a #GError, or %NULL
+ *
+ * Creates a new #RsvgHandle for @file.
+ *
+ * If @cancellable is not %NULL, then the operation can be cancelled by
+ * triggering the cancellable object from another thread. If the
+ * operation was cancelled, the error G_IO_ERROR_CANCELLED will be
+ * returned.
+ *
+ * Returns: a new #RsvgHandle on success, or %NULL with @error filled in
+ *
+ * Since: 2.32
+ */
+RsvgHandle *
+rsvg_handle_new_from_gfile_sync (GFile          *file,
+                                 RsvgHandleFlags flags,
+                                 GCancellable   *cancellable,
+                                 GError        **error)
+{
+    RsvgHandle *handle;
+    GFileInputStream *stream;
+
+    g_return_val_if_fail (G_IS_FILE (file), NULL);
+    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);
+    g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+    stream = g_file_read (file, cancellable, error);
+    if (stream == NULL)
+        return NULL;
+
+    handle = rsvg_handle_new_from_stream_sync (G_INPUT_STREAM (stream), file,
+                                               flags, cancellable, error);
+    g_object_unref (stream);
+
+    return handle;
+}
+
+/**
+ * rsvg_handle_new_from_stream_sync:
+ * @stream: a #GInputStream
+ * @base_file: (allow-none): a #GFile, or %NULL
+ * @flags: flags from #RsvgHandleFlags
+ * @cancellable: (allow-none): a #GCancellable, or %NULL
+ * @error: (allow-none): a location to store a #GError, or %NULL
+ *
+ * Creates a new #RsvgHandle for @stream.
+ *
+ * If @cancellable is not %NULL, then the operation can be cancelled by
+ * triggering the cancellable object from another thread. If the
+ * operation was cancelled, the error G_IO_ERROR_CANCELLED will be
+ * returned.
+ *
+ * Returns: a new #RsvgHandle on success, or %NULL with @error filled in
+ *
+ * Since: 2.32
+ */
+RsvgHandle *
+rsvg_handle_new_from_stream_sync (GInputStream   *stream,
+                                  GFile          *base_file,
+                                  RsvgHandleFlags flags,
+                                  GCancellable    *cancellable,
+                                  GError         **error)
+{
+    RsvgHandle *handle;
+
+    g_return_val_if_fail (G_IS_INPUT_STREAM (stream), NULL);
+    g_return_val_if_fail (base_file == NULL || G_IS_FILE (base_file), NULL);
+    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);
+    g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+    handle = rsvg_handle_new ();
+
+    if (base_file)
+        rsvg_handle_set_base_gfile (handle, base_file);
+
+    if (!rsvg_handle_read_stream_sync (handle, stream, cancellable, error)) {
+        g_object_unref (handle);
+        return NULL;
+    }
+
+    return handle;
+}
+
 /**
  * rsvg_init:
  *
@@ -1620,7 +2014,7 @@ rsvg_init (void)
 {
     g_type_init ();
 
-#ifdef HAVE_SVGZ
+#ifdef HAVE_GSF
     gsf_init ();
 #endif
 
@@ -1636,7 +2030,7 @@ rsvg_init (void)
 void
 rsvg_term (void)
 {
-#ifdef HAVE_SVGZ
+#ifdef HAVE_GSF
     gsf_shutdown ();
 #endif
 
@@ -1720,15 +2114,12 @@ rsvg_bbox_insert (RsvgBbox * dst, RsvgBbox * src)
     if (src->virgin)
         return;
 
-    if (!dst->virgin)
-        {
-            xmin = dst->x, ymin = dst->y;
-            xmax = dst->x + dst->w, ymax = dst->y + dst->h;
-        }
-    else
-        {
-            xmin = ymin = xmax = ymax = 0;
-        }
+    if (!dst->virgin) {
+        xmin = dst->x, ymin = dst->y;
+        xmax = dst->x + dst->w, ymax = dst->y + dst->h;
+    } else {
+        xmin = ymin = xmax = ymax = 0;
+    }
 
     _rsvg_affine_invert (affine, dst->affine);
     _rsvg_affine_multiply (affine, src->affine, affine);
@@ -1771,15 +2162,12 @@ rsvg_bbox_clip (RsvgBbox * dst, RsvgBbox * src)
     if (src->virgin)
         return;
 
-	if (!dst->virgin)
-		{
-			xmin = dst->x + dst->w, ymin = dst->y + dst->h;
-			xmax = dst->x, ymax = dst->y;
-		}
-	else
-		{
-			xmin = ymin = xmax = ymax = 0;
-		}
+	if (!dst->virgin) {
+        xmin = dst->x + dst->w, ymin = dst->y + dst->h;
+        xmax = dst->x, ymax = dst->y;
+    } else {
+        xmin = ymin = xmax = ymax = 0;
+    }
 
     _rsvg_affine_invert (affine, dst->affine);
     _rsvg_affine_multiply (affine, src->affine, affine);
