@@ -1,4 +1,5 @@
-/* vim: set sw=4: -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* vim: set sw=4 sts=4 ts=4 expandtab: */
 /*
    rsvg-text.h: Text handling routines for RSVG
 
@@ -29,13 +30,13 @@
 #include "rsvg.h"
 #include "rsvg-shapes.h"
 
-G_BEGIN_DECLS
+G_BEGIN_DECLS 
 
-RsvgNode * rsvg_new_text (void);
-RsvgNode * rsvg_new_tspan (void);
-RsvgNode * rsvg_new_tref (void);
-char * rsvg_make_valid_utf8 (const char *str);
+RsvgNode    *rsvg_new_text	    (void);
+RsvgNode    *rsvg_new_tspan	    (void);
+RsvgNode    *rsvg_new_tref	    (void);
+char	    *rsvg_make_valid_utf8   (const char *str, int len);
 
 G_END_DECLS
 
-#endif /* RSVG_TEXT_H */
+#endif                          /* RSVG_TEXT_H */
